/*
 * *************************************************************************************************************
 * SSD2805.c
 * MIPI Bridge Driver
 *  Created on: 08 May 2014
 *      Author: John
* *************************************************************************************************************
 */
//New1BPP and 4BPP code on 20 May 2014

#include "SSD2805_S6D04D2.h"


// ************************************Local Functions*************************************
static  void DisplayWriteReg(uint8_t ui8Reg, uint16_t ui16Data);
static  void DisplayWriteCmd(uint8_t ui8Cmd);
static  void DisplayReadReg(uint8_t ui8Reg, uint8_t *ui8Buffer);
static  void DisplaySetAddressRect(uint16_t left, uint16_t top, uint16_t right, uint16_t bottom);
static  void DisplayPutImage1BPP(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data);
static  void DisplayPutImage4BPP(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data);
static  void DisplayPutImage8BPP(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data);


enum {CLIP_X=0, CLIP_Y, CLIP_XY, CLIP_NONE};

static uint32_t ui32BackgroundColor;
static uint32_t ui32ForegroundColor;

void DisplayBackgroundSet(uint32_t ui32Color)
{
	ui32BackgroundColor = ui32Color;
}

uint32_t DisplayBackgroundGet(void)
{
	return (ui32BackgroundColor);
}

void DisplayForegroundSet(uint32_t ui32Color)
{
	ui32ForegroundColor = ui32Color;
}

uint32_t ForeBackgroundGet(void)
{
	return (ui32ForegroundColor);
}

//Set the number of data bytes (in byte) to be transmitted by SSD2805 in the next operation.
static  void SSD2805WriteTDCSize(uint32_t ui32TDCSize);

static void SSD2805WriteTDCSize(uint32_t ui32TDCSize)
{
	DisplayWriteReg(MCS_PSCR1, (uint16_t)ui32TDCSize);
	DisplayWriteReg(MCS_PSCR2, (uint16_t)(ui32TDCSize>>16));
}

/*
 * **************************************************************************************************
 * Function : 		DisplayWriteReg(uint8_t ui8Reg, uint16_t ui16Data)
 * Description : 	Local function to write a 16-bit data to the register at ui8Reg
 * Remarks: 		Speed of writing defined in the low level function in SSD2805.h with DisplayWrite()
 * 					& DisplayWrite16().
 * **************************************************************************************************
 */
static void DisplayWriteReg(uint8_t ui8Reg, uint16_t ui16Data)
{
        DisplayEnable();        //CS# low

        DisplaySetCommand();    //DC low to set command
        DisplayWrite(ui8Reg);

        DisplaySetData();       //DC high to send data

        DisplayWrite((uint8_t)(ui16Data));	//send lower byte first
        DisplayWrite((uint8_t)(ui16Data>>8));

        DisplayDisable();       //CS# high
}

/*
 * **************************************************************************************************
 * Function : 		DisplayWriteCmd(uint8_t ui8Cmd)
 * Description : 	Local function to write a 8-bit command
 * Remarks: 		Speed of writing defined in the low level function in SSD2805.h with DisplayWrite().
 *
 * **************************************************************************************************
 */
static  void DisplayWriteCmd(uint8_t ui8Cmd)
{
    DisplayEnable();        //CS# low

    DisplaySetCommand();    //DC low to set command
    DisplayWrite(ui8Cmd);

    DisplayDisable();       //CS# high
}

/*
 * ***************************************************************************************************
 * Function : void DisplaySetAddressRect(arg)
 * Description : Column Address Set and Page Address Set refer to command 0x2A and 0x2B, respectively.
 * 				 This function sets SC, EC, SP, and EP.
 * 				 Range of parameters is inclusive. That means the value of right = 239 and bottom = 239
 * 				 for a 240*240 panel. In the other words, if a picture of size 120x120 is displayed
 * 				 starting from (0,0), the coordinates for lower bottom is (119,119).
 * ***************************************************************************************************
 */
static  void DisplaySetAddressRect(uint16_t left, uint16_t top, uint16_t right, uint16_t bottom)
{
	if((left>right) || (top>bottom)) return;

	SSD2805WriteTDCSize(4);

    DisplayEnable();
    DisplaySetCommand();
    DisplayWrite(DCS_CASET);  //DCS command to MIPI slave, Column address set
    DisplaySetData();
    DisplayWrite((uint8_t)(left>>8));	//SC[15:8]
    DisplayWrite((uint8_t)left);		//SC[7:0]
    DisplayWrite((uint8_t)(right>>8));  //EC[15:8]
    DisplayWrite((uint8_t)right);   	//EC[7:0]

    DisplaySetCommand();        //DCS command to MIPI slave, Page address set
    DisplayWrite(DCS_PASET);
    DisplaySetData();
    DisplayWrite((uint8_t)(top>>8));            //SP[15:8]
    DisplayWrite((uint8_t)top);             	//SP[7:0]
    DisplayWrite((uint8_t)(bottom>>8));        	//EP[15:8]
    DisplayWrite((uint8_t)bottom);   			//EP[7:0]
    DisplayDisable();
}

static void pfnPixelDraw			(void *pvDisplayData, int32_t i32X, int32_t i32Y, uint32_t ui32Value);

static void pfnPixelDrawMultiple	(void *pvDisplayData, int32_t i32X, int32_t i32Y, int32_t i32X0,
        							int32_t i32Count, int32_t i32BPP,
        							const uint8_t *pui8Data,
        							const uint8_t *pui8Palette);

static void pfnLineDrawH			(void *pvDisplayData, int32_t i32X1, int32_t i32X2, int32_t i32Y, uint32_t ui32Value);
static void pfnLineDrawV			(void *pvDisplayData, int32_t i32X, int32_t i32Y1, int32_t i32Y2, uint32_t ui32Value);
static void pfnRectFill 			(void *pvDisplayData, const tRectangle *psRect, uint32_t ui32Value);
static uint32_t pfnColorTranslate	(void *pvDisplayData, uint32_t ui32Value);
static void pfnFlush				(void *pvDisplayData);

const tDisplay g_DisplaySolomonSSD2805 =
{
		sizeof(tDisplay),
		0,
		DISP_HOR_RESOLUTION,
		DISP_VER_RESOLUTION,
	    pfnPixelDraw,
	    pfnPixelDrawMultiple,
	    pfnLineDrawH,
	    pfnLineDrawV,
	    pfnRectFill,
	    pfnColorTranslate,
	    pfnFlush
};

static void pfnPixelDraw(	void *pvDisplayData, int32_t i32X, int32_t i32Y,
							uint32_t ui32Value)
{
	if((i32X > (DISP_HOR_RESOLUTION-1)) || (i32Y > (DISP_VER_RESOLUTION-1)) ) return;

    DisplaySetAddressRect((uint16_t)i32X,(uint16_t)i32Y, DISP_HOR_RESOLUTION-1, DISP_VER_RESOLUTION-1);

    //move to i32X0, i32Y, do we really need this???
    //SSD2805WriteTDCSize(0);
    //DisplayWriteReg(DCS_RAMWR, 0x0000);

    SSD2805WriteTDCSize(2);    //write 2 to indicate a single pixel of 2bytes in 16BPP

    DisplayEnable();
    DisplaySetCommand();
    DisplayWrite(DCS_RAMWR);
    DisplaySetData();
    DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32Value));
    DisplayDisable();

}

/*
 * ************************************************************************************************
 * Function : 	pfnPixelDrawMultiple(arg)
 * Description: This function draws multiple pixels (i.e. block of data) on display.
 * Argument:	int32_t i32X, i32Y	- top left coordinates
 * 				int32_t i32X0 		- indicates the sub-pixel offset within the pixel data for
 * 									  where the first pixel to draw in 1bpp & 4bpp
 * 				int32_t i32Count	- number of pixels to draw
 * 				unt32_t i32BPP		- bit per pixel (color depth in other words)
 * 									  It is important to notice that, i32BPP must be 1,4 or 8 for
 * 									  grlib. Palette is always used for true color in case when
 * 									  a color depth >8 is required.
 * 				const uint8_t *pui8Data	- pointer to data
 * 				const uint8_t *pui8Palette - pointer to palette data
 * ************************************************************************************************
 */
static void pfnPixelDrawMultiple (void *pvDisplayData, int32_t i32X, int32_t i32Y, int32_t i32X0,
        						  int32_t i32Count, int32_t i32BPP,
        						  const uint8_t *pui8Data,
        						  const uint8_t *pui8Palette)
{

}

//bug here when the image width is not a multiple of 8
static  void DisplayPutImage1BPP(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data)
{
 	uint8_t *pui8Palette;				//pointer to palette data
	uint16_t ui16SizeX, ui16SizeY;		//size of picture in ui16X * ui16Y
	uint8_t ui8PaletteEntry;			//maximum color index in palette (index starts from 0)
	uint32_t ui32PixelData;				//it is actually the palette entry in 24-bit for 8BPP
	uint32_t ui32X0, ui32Temp, ui32Y0;
	uint8_t ui8X0, ui8Reminder, ui8Clipping, ui8Patch = false;

	if((i32X > (DISP_HOR_RESOLUTION-1)) || (i32Y > (DISP_VER_RESOLUTION-1)) ) return;

	if(*((uint8_t *)pui8Data)!=IMAGE_FMT_1BPP_UNCOMP) return;	//just for uncompressed 1BPP for this
	pui8Data++;

	ui16SizeX = *((uint16_t *)pui8Data);
	pui8Data +=2;

	ui16SizeY = *((uint16_t *)pui8Data);
	pui8Data +=2;

	ui8PaletteEntry=  *((uint8_t *)pui8Data);
	pui8Data++;

	pui8Palette = (uint8_t *)pui8Data;			//initialize pointer to start of palette

	pui8Data += (ui8PaletteEntry+1)*3;			//increment address to start of pixel

    if((i32X+ui16SizeX) > DISP_HOR_RESOLUTION)
    {
        ui32X0 = DISP_HOR_RESOLUTION-i32X;
        if((i32Y+ui16SizeY) > DISP_VER_RESOLUTION)
            {ui8Clipping = CLIP_XY; ui32Y0 = DISP_VER_RESOLUTION-i32Y;}
        else
            {ui8Clipping = CLIP_X; ui32Y0 = ui16SizeY;}
    }
    else
    {
        ui32X0 = ui16SizeX;
        if((i32Y+ui16SizeY) > DISP_VER_RESOLUTION)
            {ui8Clipping = CLIP_Y; ui32Y0 = DISP_VER_RESOLUTION-i32Y; }
        else
            {ui8Clipping = CLIP_NONE; ui32Y0 = ui16SizeY;}
    }

    //order of DisplaySetAddressRect & set TDC size matters. They should come before ui32X0 = ui32X0/8 below
    DisplaySetAddressRect(i32X, i32Y, i32X+ui32X0-1, i32Y+ui32Y0-1);
    SSD2805WriteTDCSize(ui32X0*ui32Y0*2);

    if(ui32X0%8)
    {
    	ui8Patch = true;
    	ui8Reminder = (uint8_t) (ui32X0%8);
    }

    ui32X0 = ui32X0/8;

    ui32Temp = ui32X0;	//save ui32X0

	DisplayEnable();
    DisplaySetCommand();
	DisplayWrite(DCS_RAMWR);
	DisplaySetData();

    while(ui32Y0--)
    {
    	ui32X0 = ui32Temp;	//restore ui32X0

		   while(ui32X0--)
		   {
					//get 1bit at a time and extract the corresponding entry from palette
		    		for(ui8X0=0; ui8X0<8; ui8X0++)
		    		{
		    			ui32PixelData = ((*pui8Data >> (7-ui8X0))&0x01)*3;
		    			ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
		    			if((ui32PixelData&0x00ffffff)!=MASK)
		    			{
		    				DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));
		    			}
		    			else
		    			{
		    				DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(DisplayBackgroundGet()));
		    			}
		    		}
					pui8Data++;
            }

			if(ui8Patch==true)  //patch the last bit(s)
			{
                //get 1 bit at a time and extract the corresponding entry from palette
                for(ui8X0=0; ui8X0<ui8Reminder; ui8X0++)
                {
		    			ui32PixelData = ((*pui8Data >> (7-ui8X0))&0x01)*3;
		    			ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
		    			if((ui32PixelData&0x00ffffff)!=MASK)
		    			{
		    				DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));
		    			}
		    			else
		    			{
		    				DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(DisplayBackgroundGet()));
		    			}
                }
			}

			if((ui8Clipping==CLIP_X) || (ui8Clipping==CLIP_XY))
			{
				//offset here
				pui8Data +=(ui16SizeX-(DISP_HOR_RESOLUTION-i32X)+8-1)/8;
			}
    }

    DisplayDisable();
}


static void DisplayPutImage4BPP(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data)
{
	uint8_t *pui8Palette;				//pointer to palette data
	uint16_t ui16SizeX, ui16SizeY;		//size of picture in ui16X * ui16Y
	uint8_t ui8PaletteEntry;			//maximum color index in palette (index starts from 0)
	uint32_t ui32PixelData;				//it is actually the palette entry in 24-bit for 8BPP
	uint32_t ui32X0, ui32Temp, ui32Y0;
	uint8_t ui8Clipping, ui8Patch = false;

	if((i32X > (DISP_HOR_RESOLUTION-1)) || (i32Y > (DISP_VER_RESOLUTION-1)) ) return;

	if(*((uint8_t *)pui8Data)!=IMAGE_FMT_4BPP_UNCOMP) return;	//just for uncompressed 4BPP for this
	pui8Data++;

	ui16SizeX = *((uint16_t *)pui8Data);
	pui8Data +=2;

	ui16SizeY = *((uint16_t *)pui8Data);
	pui8Data +=2;

	ui8PaletteEntry=  *((uint8_t *)pui8Data);
	pui8Data++;

	pui8Palette = (uint8_t *)pui8Data;			//initialize pointer to start of palette

	pui8Data += (ui8PaletteEntry+1)*3;			//increment address to start of pixel

    if((i32X+ui16SizeX) > DISP_HOR_RESOLUTION)
    {
        ui32X0 = DISP_HOR_RESOLUTION-i32X;
        if((i32Y+ui16SizeY) > DISP_VER_RESOLUTION)
            {ui8Clipping = CLIP_XY; ui32Y0 = DISP_VER_RESOLUTION-i32Y;}
        else
            {ui8Clipping = CLIP_X; ui32Y0 = ui16SizeY;}
    }
    else
    {
        ui32X0 = ui16SizeX;
        if((i32Y+ui16SizeY) > DISP_VER_RESOLUTION)
            {ui8Clipping = CLIP_Y; ui32Y0 = DISP_VER_RESOLUTION-i32Y; }
        else
            {ui8Clipping = CLIP_NONE; ui32Y0 = ui16SizeY;}
    }

    //order of DisplaySetAddressRect & set TDC size matters. They should come before ui32X0 = (ui32X0-1)/2 below
    DisplaySetAddressRect(i32X, i32Y, i32X+ui32X0-1, i32Y+ui32Y0-1);
    SSD2805WriteTDCSize(ui32X0*ui32Y0*2);

    if(ui32X0%2)
    	ui8Patch = true;

    ui32X0 = ui32X0/2;

    ui32Temp = ui32X0;	//save ui32X0

	DisplayEnable();
    DisplaySetCommand();
	DisplayWrite(DCS_RAMWR);
	DisplaySetData();

    while(ui32Y0--)
    {
    	ui32X0 = ui32Temp;	//restore ui32X0

		   while(ui32X0--)
		   {
				//Get the upper nibble and extract the corresponding entry from palette
				ui32PixelData = (*pui8Data >> 4)*3;
				ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
				// Translate this palette entry and write it to the screen.
				if((ui32PixelData&0x00ffffff)!=MASK)
				{
					DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));
				}
				else
				{
					DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(DisplayBackgroundGet()));
				}

				//Get the lower nibble
				ui32PixelData = ((*pui8Data++) & 15) * 3;
				ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
				// Translate this palette entry and write it to the screen.
				if((ui32PixelData&0x00ffffff)!=MASK)
				{
					DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));
				}
				else
				{
					DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(DisplayBackgroundGet()));
				}
			}

			if(ui8Patch==true)  //patch the last nibble for each line here
			{
				ui32PixelData = (*pui8Data >> 4)*3;
				ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
				// Translate this palette entry and write it to the screen.
				if((ui32PixelData&0x00ffffff)!=MASK)
				{
					DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));
				}
				else
				{
					DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(DisplayBackgroundGet()));
				}

				pui8Data++;
			}

			if((ui8Clipping==CLIP_X) || (ui8Clipping==CLIP_XY))
			{
				//offset here
				if(ui8Patch==true)
					pui8Data +=((ui16SizeX+i32X-1)-DISP_HOR_RESOLUTION+2-1)/2;
				else
					pui8Data +=(ui16SizeX+i32X-DISP_HOR_RESOLUTION+2-1)/2;
			}
    }

    DisplayDisable();
}

/*
********************************************************************************************************
* Function:     DisplayPutImage4BPP(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data)
* Description:  This function displays an image of color depth 4BPP with special considerations
*               (1) The whole image fits in the display
*               (2) Image width in even (2,4,6, 8 etc) or odd (1,3,5,7 etc) pixels
*               (3) Only part of the image fits in the display; thus, clipping is supported
*               (4) Combination of the three conditions above
*               It is more complicated as data in Flash is arranged in nibbles for 4BPP,
*               (1 byte = 2 nibbles = 2 pixels) rather than 1 pixel per byte. Pixel data is written to
*               display in block rather than pixel-by-pixel with the function PutPixel(). It turns out
*               that speed of writing can be boosted and hence a smoother display.
*
********************************************************************************************************
*/
static  void DisplayPutImage8BPP(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data)
{
	uint8_t *pui8Palette;				//pointer to palette data
	uint16_t ui16SizeX, ui16SizeY;		//size of picture in ui16X * ui16Y
	uint8_t ui8PaletteEntry;			//maximum color index in palette (index starts from 0)
	uint32_t ui32PixelData;				//it is actually the palette entry in 24-bit for 8BPP
	uint32_t ui32X0, ui32Temp, ui32Y0;
	uint8_t ui8Clipping;

	if((i32X > (DISP_HOR_RESOLUTION-1)) || (i32Y > (DISP_VER_RESOLUTION-1)) ) return;

	if(*((uint8_t *)pui8Data)!=IMAGE_FMT_8BPP_UNCOMP) return;	//just for uncompressed 8BPP for this
	pui8Data++;

	ui16SizeX = *((uint16_t *)pui8Data);
	pui8Data +=2;
	ui16SizeY = *((uint16_t *)pui8Data);
	pui8Data +=2;

	ui8PaletteEntry=  *((uint8_t *)pui8Data);
	pui8Data++;

	pui8Palette = (uint8_t *)pui8Data;			//initialize pointer to start of palette

	pui8Data += (ui8PaletteEntry+1)*3;			//increment address to start of pixel

    if((i32X+ui16SizeX) > DISP_HOR_RESOLUTION)
    {
        ui32X0 = DISP_HOR_RESOLUTION-i32X;
        if((i32Y+ui16SizeY) > DISP_VER_RESOLUTION)
            {ui8Clipping = CLIP_XY; ui32Y0 = DISP_VER_RESOLUTION-i32Y;}
        else
            {ui8Clipping = CLIP_X; ui32Y0 = ui16SizeY;}
    }
    else
    {
        ui32X0 = ui16SizeX;
        if((i32Y+ui16SizeY) > DISP_VER_RESOLUTION)
            {ui8Clipping = CLIP_Y; ui32Y0 = DISP_VER_RESOLUTION-i32Y; }
        else
            {ui8Clipping = CLIP_NONE; ui32Y0 = ui16SizeY;}
    }


    DisplaySetAddressRect(i32X, i32Y, i32X+ui32X0-1, i32Y+ui32Y0-1);

    SSD2805WriteTDCSize(ui32X0*ui32Y0*2);
    ui32Temp = ui32X0;
    DisplayEnable();
    DisplaySetCommand();
    DisplayWrite(DCS_RAMWR);
    DisplaySetData();
    while(ui32Y0--)
    {
    	ui32X0 = ui32Temp;	//restore ui32X0 as we are counting ui32X0 below
    	while(ui32X0--)
    	{
			//Get the byte from Flash and extract the corresponding entry from palette
			ui32PixelData = (*pui8Data++)*3;
			ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
			// Translate this palette entry and write it to the screen.
			if((ui32PixelData&0x00ffffff)!=MASK)
			{
				DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));
			}
			else
			{
				DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(DisplayBackgroundGet()));
			}
		}

			if((ui8Clipping==CLIP_X) || (ui8Clipping==CLIP_XY))
			{
				//offset here
				pui8Data +=ui16SizeX+i32X-DISP_HOR_RESOLUTION;
			}
    }
    DisplayDisable();


/*
	uint8_t *pui8Palette;				//pointer to palette data
	uint16_t ui16SizeX, ui16SizeY;		//size of picture in ui16X * ui16Y
	uint8_t ui8PaletteEntry;			//maximum color index in palette (index starts from 0)
	uint32_t ui32PixelCount;
	uint32_t ui32PixelData;				//it is actually the palette entry in 24-bit for 8BPP
	uint16_t x,y;

	if((i32X > (DISP_HOR_RESOLUTION-1)) || (i32Y > (DISP_VER_RESOLUTION-1)) ) return;

	if(*((uint8_t *)pui8Data)!=IMAGE_FMT_8BPP_UNCOMP) return;	//just for uncompressed 8BPP for this
	pui8Data++;

	ui16SizeX = *((uint16_t *)pui8Data);
	pui8Data +=2;
	ui16SizeY = *((uint16_t *)pui8Data);
	pui8Data +=2;

	ui8PaletteEntry=  *((uint8_t *)pui8Data);
	pui8Data++;

	pui8Palette = (uint8_t *)pui8Data;			//initialize pointer to start of palette

	pui8Data += (ui8PaletteEntry+1)*3;			//increment address to start of pixel

	ui32PixelCount = (uint32_t)ui16SizeX*ui16SizeY;

		if(((i32X+ui16SizeX) > DISP_HOR_RESOLUTION) || ((i32Y+ui16SizeY) > DISP_VER_RESOLUTION))
		{
			//clipping handled here
			for(y=0; y < ui16SizeY; y++)
			{
				if((i32Y+y)>(DISP_VER_RESOLUTION-1)) return;	//image out of display area

				DisplaySetAddressRect(i32X, i32Y+y, DISP_HOR_RESOLUTION-1, DISP_VER_RESOLUTION-1);

				SSD2805WriteTDCSize((DISP_HOR_RESOLUTION-i32X)*2);	//*2 required for 16-bit BPP

                                DisplayEnable();
				DisplaySetCommand();
				DisplayWrite(DCS_RAMWR);
				DisplaySetData();

                                for(x=0; x<(DISP_HOR_RESOLUTION-i32X); x++)
                                {
                                        // Get the next byte of pixel data and extract the
                                        // corresponding entry from the palette.
										ui32PixelData = (*pui8Data++)*3;
										ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
                                        // Translate this palette entry and write it to the screen.
                                        if((ui32PixelData&0x00ffffff)!=MASK)
                                            DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));

                                }

                                pui8Data += (ui16SizeX+i32X-DISP_HOR_RESOLUTION);


				DisplayDisable();
			}
		}
		else
		{
			//whole image within screen boundary, optimized write possible here
			//define area of writing rectangle
			DisplaySetAddressRect((uint16_t)i32X, (uint16_t)i32Y, (uint16_t)(i32X+ui16SizeX-1), (uint16_t)(i32Y+ui16SizeY-1));

			SSD2805WriteTDCSize(0);
		    DisplayWriteReg(DCS_RAMWR, 0x0000);		//dummy write to reset Column and Page addresses

			//It is important to notice that, pixel format is NOT equivalent to data bus width
			//Data bus width in Tiva C EVK is in 8-bit width DATA[7:0] interfacing SSD2805 D[7:0].
			//However, pixel format represents the bit-per-pixel format in color space
			//In this case, the byte count = pixel count *2 since bit-per-pixel = 16BPP with 8-bit interface (hardware) and the
			//number of byte transmitted is therefore pixel count *2
		    SSD2805WriteTDCSize(ui32PixelCount*2);

		    DisplayEnable();
			DisplaySetCommand();
			DisplayWrite(DCS_RAMWRC);
			DisplaySetData();
		    //block write data to TFT
			while(ui32PixelCount--)
			{
	            // Get the next byte of pixel data and extract the
	            // corresponding entry from the palette.
				ui32PixelData = *pui8Data++*3;
				ui32PixelData = *(uint32_t *)(pui8Palette + ui32PixelData);
	            // Translate this palette entry and write it to the screen.
                if((ui32PixelData&0x00ffffff)!=MASK)
                	DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32PixelData));

			}
			DisplayDisable();
		}
*/
}

void DisplayPutImage(int32_t i32X, int32_t i32Y, const uint8_t *pui8Data)
{
	uint8_t ui8FMT;								//picture format (IMAGE_FMT_8BPP_UNCOMP defined in grlib.h, etc)


	if((i32X > (DISP_HOR_RESOLUTION-1)) || (i32Y > (DISP_VER_RESOLUTION-1)) ) return;

	ui8FMT = *((uint8_t *)pui8Data);			//get the format (1BPP, 4BPP, or 8BPP)
	switch (ui8FMT)
	{
	case IMAGE_FMT_1BPP_UNCOMP:
		DisplayPutImage1BPP(i32X, i32Y, pui8Data);
		break;
	case IMAGE_FMT_4BPP_UNCOMP:
		DisplayPutImage4BPP(i32X, i32Y, pui8Data);
		break;
	case IMAGE_FMT_8BPP_UNCOMP:
		DisplayPutImage8BPP(i32X, i32Y, pui8Data);
		break;
	}
}

static void pfnLineDrawH (void *pvDisplayData, int32_t i32X1, int32_t i32X2, int32_t i32Y, uint32_t ui32Value)
{
	uint16_t left, right, i;

	if(i32X1 < i32X2)
	{
		right=(uint16_t)i32X2;
		left=(uint16_t)i32X1;
	}
	else if (i32X2 < i32X1)
	{
		right=(uint16_t)i32X1;
		left=(uint16_t)i32X2;
	}
	else
	{
		return;
	}

    DisplaySetAddressRect(left, (uint16_t)i32Y, right, (uint16_t)i32Y);

    //SSD2805WriteTDCSize(0);
    //DisplayWriteReg(DCS_RAMWR, 0x0000);  	//dummy write to reset Start Column (left) and Start Page(i32Y) registers

    SSD2805WriteTDCSize((right-left+1)*2);	//+1 required as coordinates are all inclusive

    DisplayEnable();
    DisplaySetCommand();
    DisplayWrite(DCS_RAMWR);
    DisplaySetData();

    for(i=0; i<(right-left+1); i++)
    {
    	DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32Value));
    }

    DisplayDisable();

}



static void pfnLineDrawV (void *pvDisplayData, int32_t i32X, int32_t i32Y1, int32_t i32Y2, uint32_t ui32Value)
{
	uint16_t top, bottom, i;

	if(i32Y1 < i32Y2)
	{
		bottom=(uint16_t)i32Y2;
		top=(uint16_t)i32Y1;
	}
	else if (i32Y2 < i32Y1)
	{
		bottom=(uint16_t)i32Y1;
		top=(uint16_t)i32Y2;
	}
	else
	{
		return;
	}

    DisplaySetAddressRect((uint16_t)i32X, top, (uint16_t)i32X, bottom);

    //SSD2805WriteTDCSize(0);
    //DisplayWriteReg(DCS_RAMWR, 0x0000);  		//dummy write to reset Start Column (left) and Start Page(i32Y) registers


    SSD2805WriteTDCSize((bottom - top + 1)*2);	//+1 required as coordinates are all inclusive

    DisplayEnable();
    DisplaySetCommand();
    DisplayWrite(DCS_RAMWR);   //0x3C, memory write continue
    DisplaySetData();

    for(i=0;i<(bottom-top+1); i++)
    {
        DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32Value));
    }

    DisplayDisable();
}


static void pfnRectFill (void *pvDisplayData, const tRectangle *psRect, uint32_t ui32Value)
{
	uint16_t top, left, right, bottom;
    uint32_t ui32X0;

	left = psRect->i16XMin;
	top = psRect->i16YMin;
	right = psRect->i16XMax;
	bottom = psRect->i16YMax;

	if((right < left) || (bottom < top))
	{
		return;
	}

    DisplaySetAddressRect(left, top, right, bottom);

    //SSD2805WriteTDCSize(0);
    //DisplayWriteReg(DCS_RAMWR, 0x0000);  		//dummy write to reset Start Column (left) and Start Page(i32Y) registers

    ui32X0 = (uint32_t)(bottom-top+1)*(right-left+1)*2; 	//Byte count = pixel count *2 in 16BPP
    SSD2805WriteTDCSize(ui32X0);

    DisplayEnable();  	//dummy read to reset Start Column (i32X) and Start Page(top) registers

    DisplaySetCommand();
    DisplayWrite(DCS_RAMWR);   //0x3C, memory write continue
    DisplaySetData();

    ui32X0 = ui32X0/2;

    while(ui32X0--)
    {
    	DisplayWrite16((uint16_t)DPYCOLORTRANSLATE(ui32Value));
    }

    DisplayDisable();
}


static uint32_t pfnColorTranslate(void *pvDisplayData, uint32_t ui32Value)
{
	return ((uint32_t)DPYCOLORTRANSLATE(ui32Value));
}

static void pfnFlush(void *pvDisplayData)
{


}


//low level graphic controller reset function
void DisplayReset(void)
{
    // Enable the GPIO peripherals used to interface to the SSD1268.
    DisplayDataConfig();		//data port as output
	DisplayCmdDataConfig();		//enable DC line as output
	DisplayReadConfig();		//RD# line as output
	DisplayReadDisable();		//RD# high
    DisplayWriteConfig();		//WR# line as output
    DisplayWriteDisable();		//WR# high
    DisplayResetConfig();		//RST line
    DisplayConfig();			//CS# line

    DisplayBacklightConfig();	//backlight PWM control
    DisplayBacklightOn();

    DisplayResetDisable();
    DelayMs(300);
    DisplayResetEnable();
    DelayMs(10);
    DisplayResetDisable();
    DelayMs(10);

    //Step 1: Set PLL. Do it slow because PLL has not started. The min WRX period is 6T with T=1/20MHz
    //Therefore the min peiod for WRX = 300ns with an external TX_CLK = 20MHz
    //PLL 	= clock*MUL/(PDIV*DIV)
    //		= clock*(BAh[7:0]+1)/((BAh[15:12]+1)*(BAh[11:8]+1))
    //		= 20*(0x0f+1)/1*1 = 20*16 = 320MHz
    //Remark: 350MHz >= fvco >= 225MHz for SSD2805 since the max. speed per lane is 350Mbps
    //start DisplayWriteReg(0xba, 0x000f)
    DisplayEnable();        //CS# low

    DisplaySetCommand();    //DC low to set command
    { 	HWREG(LCD_DATAH_BASE + GPIO_O_DATA + (LCD_DATAH_PINS << 2)) = (MCS_PLCR);\
    	DisplayWriteEnable();ROM_SysCtlDelay(10); DisplayWriteDisable();
    }

    DisplaySetData();       //DC high to send data

    //important!!! SSD2805 receives lower byte first!
    {	HWREG(LCD_DATAH_BASE + GPIO_O_DATA + (LCD_DATAH_PINS << 2)) = (uint8_t)(0x0f);\
    	DisplayWriteEnable();ROM_SysCtlDelay(10);DisplayWriteDisable();\
    	HWREG(LCD_DATAH_BASE + GPIO_O_DATA + (LCD_DATAH_PINS << 2)) = (uint8_t)(0x00);\
    	DisplayWriteEnable();ROM_SysCtlDelay(10); DisplayWriteDisable();}

    DisplayDisable();       //CS# high
    //end DisplayWriteReg(0xba, 0x000f)

    //start DisplayWriteReg(0xb9, 0x0001) to enable PLL
    DisplayEnable();        //CS# low

    DisplaySetCommand();    //DC low to set command
    { 	HWREG(LCD_DATAH_BASE + GPIO_O_DATA + (LCD_DATAH_PINS << 2)) = (MCS_PCR);\
    	DisplayWriteEnable();ROM_SysCtlDelay(10); DisplayWriteDisable();
    }

    DisplaySetData();       //DC high to send data

    {	HWREG(LCD_DATAH_BASE + GPIO_O_DATA + (LCD_DATAH_PINS << 2)) = (uint8_t)(0x01);\
    	DisplayWriteEnable();ROM_SysCtlDelay(10); DisplayWriteDisable();\
    	HWREG(LCD_DATAH_BASE + GPIO_O_DATA + (LCD_DATAH_PINS << 2)) = (uint8_t)(0x00);\
    	DisplayWriteEnable();ROM_SysCtlDelay(10); DisplayWriteDisable();}

    DisplayDisable();       //CS# high
    //end DisplayWriteReg(0xb9, 0x0001)

    DelayMs(2);	//wait 2msec for PLL started


	//Step 2: Now it is safe to set PMP at max. speed. Nothing to do here


	//Step 3: set clock control register for SYS_CLK & LP clock speed
	//SYS_CLK = TX_CLK/(BBh[7:6]+1), TX_CLK = external oscillator clock speed
	//In this case, SYS_CLK = 20MHz/(1+1)=10MHz. Measure SYS_CLK pin to verify it.
	//LP clock = PLL/(8*(BBh[5:0]+1)) = 320/(8*(4+1)) = 8MHz, conform to AUO panel's spec, default LP = 8Mbps
	//S6D04D2 is the controller of AUO/LG 1.54" panel.
	DisplayWriteReg(MCS_CCR, 0x0044);
	DisplayWriteReg(MCS_TR, 0x0100);	//output sys_clk for debug. Now check sys_clk pin for 10MHz signal

	//Step 4: Set MIPI packet format
	DisplayWriteReg(MCS_CFGR, 0x0243);	//EOT packet enable, write operation, it is a DCS packet
										//HS clock is disabled, video mode disabled, in HS mode to send data

	//Step 5: set Virtual Channel (VC) to use
	DisplayWriteReg(MCS_VCR, 0x0000);


	//Step 6: Now write DCS command to AUO/LG panel for system power-on upon reset
	SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_SLPOUT,0x0000);     //DCS sleep-out command

    DelayMs(100);							//wait for AUO/LG  panel after sleep out

	//Step 7: Now configuration parameters sent to AUO/LG
    SSD2805WriteTDCSize(1);						//define TDC size to be 1
    DisplayWriteReg(DCS_COLMOD,0x0005);         //DCS command to set pixel format

    DisplayClearDevice(0x0000);

	SSD2805WriteTDCSize(0);						//define TDC size
    DisplayWriteReg(DCS_DISPON,0x0000);    	    //display ON DCS command to AUO/LG panel
}

void DisplayClearDevice(GFX_COLOR color)
{
    uint32_t ui32X0;

    ui32X0 = (uint32_t)DISP_HOR_RESOLUTION*DISP_VER_RESOLUTION*2;	//count in byte therefore we need to multiply by 2 for 16BPP

    DisplaySetAddressRect(0, 0, DISP_HOR_RESOLUTION-1, DISP_VER_RESOLUTION-1);

    SSD2805WriteTDCSize(ui32X0);
    DisplayWriteReg(MCS_PSCR3, 0x0400);					//set threshold to max size for block write

    DisplayEnable();
    DisplaySetCommand();
    DisplayWrite(DCS_RAMWR);

    DisplaySetData();

    ui32X0 = ui32X0/2;

    while(ui32X0--)
    {
        DisplayWrite(((uint16_val)color).v[0]);
        DisplayWrite(((uint16_val)color).v[1]);
    }

    DisplayDisable();    //CS# high to disable
}

void DisplayRectFill(uint16_t left, uint16_t top, uint16_t right, uint16_t bottom, uint32_t ui32Value)
{
	tRectangle *psRect;
	tRectangle sRect;

	sRect.i16XMin = left;
	sRect.i16YMin = top;
	sRect.i16XMax = right;
	sRect.i16YMax = bottom;

	psRect =&sRect;

	pfnRectFill(NULL, (const tRectangle *)psRect, ui32Value);
}

void DisplayLineDrawV (int32_t i32X, int32_t i32Y1, int32_t i32Y2, uint32_t ui32Value)
{
	pfnLineDrawV (NULL, i32X, i32Y1, i32Y2, ui32Value);
}

void DisplayLineDrawH (int32_t i32X1, int32_t i32X2, int32_t i32Y, uint32_t ui32Value)
{
	pfnLineDrawH (NULL, i32X1, i32X2, i32Y, ui32Value);
}

void DisplayOn(void)
{
    	SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_DISPON,0x0000);
}

void DisplayOff(void)
{
    	SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_DISPOFF,0x0000);
}

void DisplaySleepIn(void)
{
        SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_SLPIN,0x0000);
}

void DisplaySleepOut(void)
{
        SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_SLPOUT,0x0000);
}

void DisplayIdleOn(void)
{
        SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_IDMON,0x0000);
}

void DisplayIdleOff(void)
{
        SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_IDMOFF,0x0000);
}

void DisplayPutPixel(int16_t x, int16_t y, GFX_COLOR color)
{
    DisplaySetAddressRect((uint16_t)x, (uint16_t)y, DISP_HOR_RESOLUTION-1, DISP_VER_RESOLUTION-1);

    SSD2805WriteTDCSize(2);

    DisplayEnable();
    DisplaySetCommand();
    DisplayWrite(DCS_RAMWR);
    DisplaySetData();
    DisplayWrite(((uint16_val)color).v[1]);
    DisplayWrite(((uint16_val)color).v[0]);
    DisplayDisable();
}

void DisplaySoftReset(void)
{
    SSD2805WriteTDCSize(0);
    DisplayWriteReg(DCS_SWRESET,0x0000);
}
